/* png_io.c  --  Handle reading and writing of heightfields in PNG format.
 *               Output is only supported for 8/16-bit greyscale PNG type.
 *               4/28/96  John P. Beale
 */

#ifdef HAS_PNG

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <png.h>
#include "hl.h"

#define U unsigned int

extern int sigbits;         /* number of significant bits to store */

/* --------------------------------------- */

int png_write(hfield hf, char *fname)
{
  png_byte *pngrow = NULL;
  png_struct *png_ptr;
  png_info *info_ptr;
  png_byte mask_lo, onebit;   /* for masking off unused bits from byte */ 

  int xsize, ysize;
  int x, y;
  FILE *fpout;
  U color;

    /* ------------- open PNG file ----------- */

  xsize = hf.xsize;
  ysize = hf.ysize;
  fpout = fopen(fname, "wb");
  if (fpout == NULL) {
    fprintf(stderr,"Trouble opening PNG file %s\n",fname);
    return(1);
  }

 /* -------------- Write heightfield in PNG format ---------------- */

	 /* allocate the necessary structures */
     png_ptr = (png_struct *)malloc(sizeof (png_struct)); 
     if (!png_ptr)
     {
      fclose(fpout);
      printf("Error opening PNG file.\n");
      return (1);
     }
     
     /* might be 8 or 16-bit file, but extra mem. doesn't hurt */
     if ( (pngrow = (png_byte *)malloc(2*xsize*sizeof(png_byte)))
	       == NULL) {
	     fclose(fpout);
	     printf("Memory allocation error");
	     return(1);
	   }

     info_ptr = (png_info *)malloc(sizeof (png_info));
     if (!info_ptr)
     {
      fclose(fpout);
      free(png_ptr);
      printf("Memory error opening PNG file.\n");
      return(1);
     }

     /* set error handling */
     if (setjmp(png_ptr->jmpbuf))
     {
      png_write_destroy(png_ptr);
      fclose(fpout);
      free(png_ptr);
      free(info_ptr);
      /* If we get here, we had a problem reading the file */
      printf("Had error opening PNG file.\n");
      return(1);
     }

     png_info_init(info_ptr);  /* initialize the structures */
     png_write_init(png_ptr);
     png_init_io(png_ptr, fpout);     /* set up the output control */

    /* -------- set the PNG file header information here ----------- */
     info_ptr->width = xsize;
     info_ptr->height = ysize;
     info_ptr->color_type = PNG_COLOR_TYPE_GRAY;
     if (sigbits > 8) {
       info_ptr->bit_depth = 16;
       info_ptr->pixel_depth = 16;
     } else {
       info_ptr->bit_depth = 8;
       info_ptr->pixel_depth = 8;
     }
     info_ptr->sig_bit.gray = (png_byte) sigbits;  /* actual # sig. bits */
     info_ptr->compression_type = 0;
     info_ptr->filter_type = 0;
     info_ptr->interlace_type = 0; /* non-interlaced */
     info_ptr->valid |= PNG_INFO_sBIT;     /* optional significant bit chunk */
     info_ptr->valid |= PNG_INFO_gAMA;     /* optional gamma chunk */
     info_ptr->gamma = 1.0;

		       /* ----------- Write PNG header----------- */
     png_write_info(png_ptr, info_ptr);

     /* ---------  Write out PNG file line-by-line --------- */

     if (sigbits > 8) { /* write 16-bit PNG greyscale file */
      onebit = 128;   /* binary 10000000  */
      mask_lo = 0;
      for(y=0;y<(sigbits-8);y++) {  /* build up mask byte */
	mask_lo ^= onebit;
	onebit >>= 1;
      }
 
      for(y=0;y<ysize;y++) {
	for(x=0;x<xsize;x++) {
	  color = (unsigned int) (65535.999 * El(hf.a,x,y));
	  pngrow[2*x] = color >> 8;
	  pngrow[2*x + 1] = (color % 256) & mask_lo;
	}
	png_write_row(png_ptr, pngrow);
      }
     } else {  /* write 8-bit PNG greyscale file */
      onebit = 128;   /* binary 10000000  */
      mask_lo = 0;
      for(y=0;y<sigbits;y++) {  /* build up mask byte */
	mask_lo ^= onebit;
	onebit >>= 1;
      }
 
      for(y=0;y<ysize;y++) {
	for(x=0;x<xsize;x++) {
	  color = (unsigned int) (65535.999 * El(hf.a,x,y));
	  pngrow[x] = (color >> 8) & mask_lo;
	}
	png_write_row(png_ptr, pngrow);
      }

     } /* end of PNG data writing */

     /* ------- PNG file written, now write trailer & close ------ */

      png_write_end(png_ptr, info_ptr);
      png_write_destroy(png_ptr);
      free(png_ptr);
      free(info_ptr);
      fclose(fpout);
      return(0);
} /* png_write */


/* png_read() -- read a png file into a heightfield structure
 *               based on example.c from libpng 0.8
 */

int png_read(hfield *hfp, char *file_name)
{
  hfield hf;
  int x, y;
  int xsize, ysize;     /* necessary for EL() macro */
  FILE *fp;
  size_t imem;
  int ct;               /* PNG color type */
  png_struct *png_ptr;
  png_info *info_ptr;
  png_byte **rowptr;   /* array of row vectors for image storage */

   /* open the file */
   fp = fopen(file_name, "rb");
   if (!fp) {
      printf("Error opening PNG file %s.\n",file_name);
      return(1);
   }

   /* allocate the necessary structures */
   png_ptr = malloc(sizeof (png_struct));
   if (!png_ptr)
   {
      fclose(fp);
      printf("Error opening PNG file %s: out of memory.\n",file_name);
      return(1);
   }

   info_ptr = malloc(sizeof (png_info));
   if (!info_ptr)
   {
      fclose(fp);
      free(png_ptr);
      printf("Error opening PNG file %s: out of memory.\n",file_name);
      return(1);
   }

   /* set error handling */
   if (setjmp(png_ptr->jmpbuf))
   {
      png_read_destroy(png_ptr, info_ptr, (png_info *)0);
      fclose(fp);
      free(png_ptr);
      free(info_ptr);
      /* If we get here, we had a problem reading the file */
      printf("Error reading PNG file %s.\n",file_name);
      return(1);
   }

#if !defined( PNG_LIBPNG_VER ) || PNG_LIBPNG_VER < 89
   /* This is required, for some unknown reason, with libpng 0.88 */
   png_set_message_fn(png_ptr, NULL, NULL, NULL);
#endif
   
   /* initialize the structures, info first for error handling */
   png_info_init(info_ptr);
   png_read_init(png_ptr);

   /* set up the input control */
   png_init_io(png_ptr, fp);

   /* read the file information */
   png_read_info(png_ptr, info_ptr);
   ct = info_ptr->color_type;
   xsize = info_ptr->width;
   ysize = info_ptr->height;
   /*   printf("[%d x %d]\n",xsize,ysize); */

    /* expand paletted colors into true rgb */
  /*
   if (ct == PNG_COLOR_TYPE_PALETTE)
      png_set_expand(png_ptr);
  */

   /* expand sub-8-bit grayscale images to the full 8 bits */
 /*
   if ((ct == PNG_COLOR_TYPE_GRAY) && (info_ptr->bit_depth < 8))
      png_set_expand(png_ptr);
   */
   /* tell libpng to strip 16 bit depth files down to 8 bits */
   if ( (info_ptr->bit_depth == 16) &&
	(ct != PNG_COLOR_TYPE_GRAY) )
				    /* 16 bit "super-trucolor" image */
	png_set_strip_16(png_ptr);

    png_read_update_info(png_ptr,info_ptr);
    /*    printf("Rowbytes needed: %ld\n",(long)info_ptr->rowbytes); */

    if ((ct != PNG_COLOR_TYPE_GRAY) && (ct != PNG_COLOR_TYPE_RGB) &&
	(ct != PNG_COLOR_TYPE_PALETTE) ) {
      printf("Sorry, unsupported PNG file type.\n");
    } else
      {
	
	/*   if (info_ptr->valid & PNG_INFO_sBIT)
     printf("%d sig. bits  ",(int)info_ptr->sig_bit.gray);
     */

   imem = (size_t) xsize;
   if ((info_ptr->bit_depth==16) && (ct == PNG_COLOR_TYPE_GRAY))
    imem *= 2;   /* 16-bit GRAY HF */
   else if (ct == PNG_COLOR_TYPE_RGB) 
    imem *= 3;   /* 24-bit color */

   /*
  printf("allocating %ld for %d x %d HF %s...\n",(long)imem,
	     xsize,ysize,file_name);
	     */
  
  if ((rowptr = (png_byte **) malloc((size_t)ysize*sizeof(png_byte *)) ) == NULL) {
	     fclose(fp); printf("Memory allocation error\n");
	     free(png_ptr); free(info_ptr);
	     return(1);
   } 

   for (y=0;y<ysize;y++) {
/*     printf("%d",y);fflush(stdout); */
     if ((rowptr[y] = (png_byte *)malloc(imem)) == NULL) {
	     fclose(fp); printf("Memory allocation error\n");
	     for (x=0;x<y;x++) free(rowptr[x]);
	     free(rowptr);
	     free(png_ptr); free(info_ptr);
	     return(1);
     } 
   }
   /*
   printf("allocated %ld for %d x %d HF %s...reading image.\n",(long)imem,
	     xsize,ysize,file_name);
	     */

   png_read_image(png_ptr,rowptr);

   if (h_newr(&hf,xsize,ysize,file_name)) {
     printf("Out of memory reading %s.\n",file_name);
     free(png_ptr);
     free(info_ptr);
     fclose(fp);
     return(1);
   }
   
   if (ct == PNG_COLOR_TYPE_RGB) {  /* read 24-bit RGB */
     for(y=0;y<ysize;y++) {
       for(x=0;x<xsize;x++) {
	   El(hf.a,x,y) = ((PTYPE)rowptr[y][3*x])/256.0 + 
	      (rowptr[y][3*x+1])/65536.0;
       } /* for x */
     } /* for y */

   } else if (info_ptr->bit_depth < 16) {  /* read 8-bit GREY or PALETTE */
     for(y=0;y<ysize;y++) {
	  for(x=0;x<xsize;x++) {
	   El(hf.a,x,y) = ((PTYPE)rowptr[y][x])/256.0; 
	  } /* for x */
     } /* for y */
   } else {                        /* read 16-bit GREY */
     for(y=0;y<ysize;y++) {
       for(x=0;x<xsize;x++) {
	   El(hf.a,x,y) = ((PTYPE)rowptr[y][2*x])/256.0 + 
	      (rowptr[y][2*x+1])/65536.0;
       } /* for x */
     } /* for y */
   
   } /* else if (16-bit) */
   *hfp = hf;

   for (y=0;y<ysize;y++) free(rowptr[y]);
   free(rowptr);

   } /* if type = GREY,RGB,Palette */

   /* clean up after the read, and free any memory allocated */
   png_read_destroy(png_ptr, info_ptr, (png_info *)0);

   /* free the structures */
   free(png_ptr);
   free(info_ptr);

   fclose(fp);
   return(0);

} /* png_read */

#endif /* HAS_PNG */
